#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Define utility functions for FAL commands."""

import re
import fal_info
import ionas.request as request
import iobase.xmldata as xmldata
from iobase.xmldata import XmlNode
from product_conf import (STOR_IF_TO_VOL, SUDO_CMD)
from product_conf.devinfo import *
from product_conf.reqinfo import (Results, ReqTypes)

# XML Tag Name
STATE_NAME = "state"
ARGLIST_NAME = "arglist"
ARG_NAME = "arg"
LOG_NAME = "log"
EVENTINFO_NAME = "eventinfo"

# X path
STATE_XPATH = "./statelist/state"
STATELIST_NAME = "statelist"
STATE_NAME = "state"

### XML Attribute ###
NAME_ATTR = "name"
CATEGORY_ATTR = "category"
ACTION_ATTR = "action"
PREVIOUS_ATTR = "previous"

BOOTING_STATE = "booting"


STATEMGR_QUERY_XML = '<statemgr><statelist><state category="system"/>' + \
                        '</statelist></statemgr>'
STATUS_LED_BASE_REQ_XML = \
  '<led>' + \
  ' <status color=\"%s\">%s</status>' + \
  '</led>'
CLEAR_ANNOUNCE_BASE_REQ_XML = \
  '<announce>' + \
  ' <option level=\"%s\" category=\"%s\"></option>' + \
  '</announce>'

TRUE = xmldata.XmlConfNode.TRUE_TEXT
FALSE = xmldata.XmlConfNode.FALSE_TEXT


def _get_entry(type, base_xml_dict, file_name=None, altfile_name=None,
        opt_attrs=None):
    """Create entry and return it."""

    return request.Entry(type, base_xml_dict[type], file_name=file_name,
                            altfile_name=altfile_name, opt_attrs=opt_attrs)


def _request_entry(pal_cmd, entry, dry_run, debug):
    """Send "entry" to "pal_cmd" and return entry in response data."""

    return request.Request(pal_cmd).request(entry,
                                    dry_run=dry_run, debug=debug).get_entry()


def _statemgr_request(type, category, state=None,
        dry_run=None, debug=None):
    """Send request to "fal_statemgr"."""

    entry = _get_entry(type, fal_info.FAL_STATEMGR_REQ_BASE_XML)
    if type == ReqTypes.QUERY:
        entry.get().get_node(STATELIST_NAME).append_child(
                            STATE_NAME, attrs={CATEGORY_ATTR: category})
    else:
        state_node = entry.get().get_node(STATE_NAME)
        state_node.set_attrs({CATEGORY_ATTR: category})
        state_node.set_text(state)

    response_entry = _request_entry(fal_info.FAL_STATEMGR_PATH, entry,
                            dry_run, debug)

    if response_entry.get_result() == Results.CHANGE_ERROR:
        raise ChangeError
    if response_entry.get_result() != Results.OK:
        raise RequestError

    return response_entry.get()


def get_internal_vol_name():

    return STOR_IF_TO_VOL["internal"][1]


def request_entry(fal_pal_cmd, entry, filter_cmd=SUDO_CMD):
    """Send "entry" to "fal/pal_cmd" and return the entry in response data."""

    return request.Request(fal_pal_cmd, filter_cmd).request(entry).get_entry()


def get_system_state():
    """Return the state of "system"."""

    try:
        entry = request.Entry(ReqTypes.QUERY, STATEMGR_QUERY_XML)
        response_entry = request_entry(fal_info.FAL_STATEMGR_PATH, entry)

        if response_entry.get_result() != Results.OK:
            return BOOTING_STATE

        return response_entry.get().get_node(STATE_XPATH).get_text()
    except:
        return BOOTING_STATE


def save_event_log(msg_code, arg_dict={}):
    """save event log."""

    EVENTLOG_XML_TMPL = '<eventlog><log><code>%s</code></log></eventlog>'

    eventlog_node = XmlNode(EVENTLOG_XML_TMPL % msg_code)

    arglist_node = XmlNode(ARGLIST_NAME)
    for arg_name, arg_value in arg_dict.iteritems():
        arglist_node.append_child(ARG_NAME, arg_value, {NAME_ATTR: arg_name})
        eventlog_node.get_node(LOG_NAME).append_child(arglist_node)

    entry = request.Entry(ReqTypes.ADD, eventlog_node)
    response_entry = request_entry(fal_info.FAL_EVENTLOG_PATH, entry)
    return


def dispatch_event(category, action, body, previous=False,
        dry_run=None, debug=None):
    """Dispatch event by "pal_events"."""

    EVENTS_REQ_XML_BASE = '<events><eventinfo/></events>'
    entry = request.Entry(ReqTypes.EVENT, EVENTS_REQ_XML_BASE)
    attrs = {CATEGORY_ATTR: category, ACTION_ATTR: action}
    if previous:
        attrs[PREVIOUS_ATTR] = TRUE
    entry.get().get_node(EVENTINFO_NAME).set_attrs(attrs)
    entry.get().get_node(EVENTINFO_NAME).append_child(body)
    response_entry = request_entry(fal_info.FAL_EVENTS_PATH, entry)
    if response_entry.get_result() != Results.OK:
        raise EventError
    return


def is_volume_share_mode(vol_name):

    USB_PORT_ENABLE_MODE_XPATH = \
        "usblist/usb[@port='%s']/modelist/mode[@enable='true']"

    if get_internal_vol_name() == vol_name:
        return True

    match = re.match("usb([0-9]+)", vol_name)
    if not match:
        return False
    port_num = match.group(1)

    entry = request.Entry(ReqTypes.LOAD)
    port_conf_node = \
        request_entry(fal_info.FAL_PORTS_PATH, entry).get()

    try:
        mode = port_conf_node.get_node(\
                   USB_PORT_ENABLE_MODE_XPATH % port_num).get_text()
    except:
        return False

    if "share" == mode:
        return True

    return False


def statemgr_query(category, dry_run=None, debug=None):
    """Return the state of "category" by "fal_statemgr"."""

    result = _statemgr_request(ReqTypes.QUERY, category,
                    dry_run=None, debug=None)
    return result.get_node(STATE_XPATH).get_text()


def statemgr_set(category, state, dry_run=None, debug=None):
    """Set the state of "category" to "state" by "fal_statemgr"."""

    _statemgr_request(ReqTypes.SET, category, state, dry_run=None, debug=None)


def statemgr_change(category, state, dry_run=None, debug=None):
    """Change the state of "category" to "state" by "fal_statemgr"."""

    _statemgr_request(ReqTypes.CHANGE, category, state,
                        dry_run=None, debug=None)


def change_status_led(color, led_mode, dry_run=None, debug=None):
    """Change STATUS LED to coler and mode.
       "coler" can be set "grean" or "red".
       "led_mode" can be set "blink" or "on".
    """

    entry = request.Entry(ReqTypes.CONTROL,
                          STATUS_LED_BASE_REQ_XML % (color, led_mode))
    _request_entry(fal_info.FAL_LED_PATH, entry, dry_run, debug)
    return


def clear_announce(level="1", category="", dry_run=None, debug=None):
    """Clear announe to level and category.
       "level" can be set "1" or "2" or "all".
       "category" set "" is all category.
    """

    entry = request.Entry(ReqTypes.CLEAR,
                          CLEAR_ANNOUNCE_BASE_REQ_XML % (level, category))
    _request_entry(fal_info.FAL_ANNOUNCE_PATH, entry, dry_run, debug)
    return

def play_melody(melody_type):

    BUZZER_REQ_XML = "<buzzer><melody type='%s'/></buzzer>"

    entry = request.Entry(ReqTypes.PLAY, BUZZER_REQ_XML % melody_type)
    response_entry = request_entry(fal_info.FAL_BUZZER_PATH, entry)
    if response_entry.get_result() != Results.OK:
        raise RequestError

    return


#
# Public classes
#
class EventError(Exception):
    """Event error"""

    pass


class RequestError(Exception):
    """Request error"""

    pass


class ChangeError(Exception):
    """Change error"""

    pass
